#include "avisynth.h"
#include "windows.h"
#include "stdio.h"
#include "info.h"

#define BLKSIZE 32

class Dup : public GenericVideoFilter
{
	double threshold;
	bool show, copy, debug;
	int successive_dups, dup_frame, maxcopies, wasmax;
public:
    Dup(PClip _child, double _threshold, bool _show, bool _copy, int _maxcopies,
		bool _debug, IScriptEnvironment* env) :
	    GenericVideoFilter(_child), threshold(_threshold), show(_show), copy(_copy),
		maxcopies(_maxcopies), debug(_debug)
	{
		if (!vi.IsYUY2())
			env->ThrowError("Dup: requires YUY2 source; use ConvertToYUY2()");
		if (threshold < 0.0 || threshold > 100.0)
			env->ThrowError("Dup: threshold out of range (0.0-100.0)");
		if (debug)
		{
			char b[80];
			sprintf(b, "Dup %s by Donald Graft, Copyright 2002\n", "1.4");
			OutputDebugString(b);
		}
	    successive_dups = 0;
		dup_frame = 0;
		wasmax = true;
	 }
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

PVideoFrame __stdcall Dup::GetFrame(int n, IScriptEnvironment* env)
{
	int get;
	PVideoFrame fcopy;

	/* Get the current frame. */
	if (n == 0)
	{
		successive_dups = 0;
		dup_frame = 0;
		wasmax = true;
	}
    PVideoFrame src = child->GetFrame(n, env);

	/* Get the frame to compare it against. It may be the first
	   in a string of copies we are emitting. */
	if (copy == true)
	{
		if (dup_frame == 0) get = 0;
		else get = dup_frame;
	}
	else
	{
		get = (n == 0 ? 0 : n -1);
	}
    PVideoFrame src0 = child->GetFrame(get, env);
	if (debug)
	{
		char buf[80];
		sprintf(buf, "%d: comparing to %d", n, get);
		OutputDebugString(buf);
	}

    /* Get destination frame. */
	PVideoFrame dst = env->NewVideoFrame(vi);

    const unsigned char *srcp = src->GetReadPtr();
    const unsigned char *src0p = src0->GetReadPtr();
	const unsigned char *srcp_saved = srcp, *src0p_saved = src0p;
    unsigned char *dstp = dst->GetWritePtr();
	unsigned char *dstp_saved = dstp;
 
    const int src_pitch = src->GetPitch();
    const int dst_pitch = dst->GetPitch();
    const int row_size = dst->GetRowSize();
    const int height = dst->GetHeight();

	int highest_x, highest_y;
	unsigned int *sum, highest_sum;
	double metric;
	int i, j, x, y;
	int xblocks, yblocks;
	char buf[80];
	int xlim, ylim, xtmp, ytmp;

	xblocks = row_size / (2*BLKSIZE);
	if (xblocks % (2*BLKSIZE)) xblocks++;
	yblocks = height / BLKSIZE;
	if (yblocks % BLKSIZE) yblocks++;
	sum = (unsigned int *) malloc(xblocks * yblocks * sizeof(unsigned int));
	if (sum == NULL)
		env->ThrowError("Dup: cannot allocate needed memory");

	srcp = srcp_saved;
	src0p = src0p_saved;
	dstp = dstp_saved;
 	for (i = 0; i < yblocks; i++)
	{
 		for (j = 0; j < xblocks; j++)
		{
			sum[i*xblocks+j] = 0;
		}
	}
	for (y = 0; y < height; y++)
	{
        for (x = 0; x < row_size; x+=2)
		{
			dstp[x] = srcp[x];
			dstp[x+1] = srcp[x+1];
			sum[(y/BLKSIZE)*xblocks + x/(2*BLKSIZE)] += abs((int)srcp[x] - (int)src0p[x]);
		}
        srcp += src_pitch;
        src0p += src_pitch;
		dstp += dst_pitch;
    }
	highest_sum = 0;
	highest_x = highest_y = 0;
	for (i = 0; i < yblocks; i++)
	{
		for (j = 0; j < xblocks; j++)
		{
#if 0
			{
				sprintf(buf, "%d: x = %d y = %d, metric = %2.2f%%\n", n, j, i,
					    (sum[i * xblocks + j] * 100) / (235*BLKSIZE*BLKSIZE));
				OutputDebugString(buf);
			}
#endif
			if (sum[i * xblocks + j] > highest_sum)
			{
				highest_sum = sum[i * xblocks + j];
				highest_x = j * BLKSIZE * 2;
				highest_y = i * BLKSIZE;
			}
		}
	}
	metric = (highest_sum * 100.0) / (235*BLKSIZE*BLKSIZE);

	if (show == true)
	{
		xlim = highest_x + 2*BLKSIZE;
		if (xlim > row_size) xlim = row_size;
		ylim = highest_y + BLKSIZE;
		if (ylim > height) ylim = height;
		for (y = highest_y; y < ylim; y++)
		{
			(dstp_saved + y * dst_pitch)[highest_x] = 235;
			xtmp = highest_x+2*(BLKSIZE - 1);
			if (xtmp < row_size)
				(dstp_saved + y * dst_pitch)[xtmp] = 235;
		}
		for (x = highest_x; x < xlim; x+=4)
		{
			(dstp_saved + (highest_y) * dst_pitch)[x] = 235;
			(dstp_saved + (highest_y) * dst_pitch)[x+2] = 235;
			(dstp_saved + (highest_y) * dst_pitch)[x+1] = 128;
			(dstp_saved + (highest_y) * dst_pitch)[x+3] = 128;
			ytmp = highest_y + BLKSIZE - 1;
			if (ytmp < height)
			{
				(dstp_saved + ytmp * dst_pitch)[x] = 235;
				(dstp_saved + ytmp * dst_pitch)[x+2] = 235;
				(dstp_saved + ytmp * dst_pitch)[x+1] = 128;
				(dstp_saved + ytmp * dst_pitch)[x+3] = 128;
			}
		}
		if (metric < threshold)
		{
			for (y = highest_y, x = 0; y < ylim; y++, x++)
			{
				xtmp = highest_x+2*x;
				if (xtmp < row_size)
					(dstp_saved + y * dst_pitch)[highest_x+2*x] = 235;
				xtmp = highest_x+2*(BLKSIZE - 1 - x);
				if (xtmp < row_size)
					(dstp_saved + y * dst_pitch)[highest_x+2*(BLKSIZE - 1 - x)] = 235;
			}
		}
		sprintf(buf, "frm %d: highest diff %2.2f%%", n, metric);
		DrawString(dst, 0, 0, buf);
	}
	if (debug)
	{
		sprintf(buf, "%d: metric = %2.2f%%\n", n, metric);
		OutputDebugString(buf);
	}

	if (copy == true)
	{
		if (metric < threshold && successive_dups < maxcopies && wasmax == false)
		{
			successive_dups++;
			if (successive_dups >= maxcopies) wasmax = true;
			if (debug)
			{
				sprintf(buf, "%d: %d successive dups, duplicating frame %d\n",
						n, successive_dups, dup_frame);
				OutputDebugString(buf);
			}
			get = dup_frame;
			fcopy = child->GetFrame(get, env);
			if (show == true)
			{
				env->MakeWritable(&fcopy);
				sprintf(buf, "frm %d: highest diff %2.2f%%", n, metric);
				DrawString(fcopy, 0, 0, buf);
				sprintf(buf, "%d successive dups", successive_dups);
				DrawString(fcopy, 0, 1, buf);
				sprintf(buf, "Copying frame %d", dup_frame);
				DrawString(fcopy, 0, 2, buf);
			}
			free(sum);
			return fcopy;
		}
		else
		{
			successive_dups = 0;
			dup_frame = n;
			wasmax = false;
		}
	}
	free(sum);
    return dst;
}

AVSValue __cdecl Create_Dup(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	char path[1024];
	char buf[80], *p;
	double  threshold = 3;
	bool show = false;
	bool copy = false;
	int  maxcopies = 10;
	bool debug = false;

	/* Load user defaults if they exist. */
	try
	{
		FILE *f;

		const char* plugin_dir = env->GetVar("$PluginDir$").AsString();
		strcpy(path, plugin_dir);
		strcat(path, "\\Dup.def");
		if ((f = fopen(path, "r")) != NULL)
		{
			while(fgets(buf, 80, f) != 0)
			{
				if (strncmp(buf, "threshold=", 10) == 0)
				{
					p = buf;
					while(*p++ != '=');
					threshold = atof(p);
				}
				if (strncmp(buf, "show=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') show = true;
					else show = false;
				}
				if (strncmp(buf, "copy=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') copy = true;
					else copy = false;
				}
				if (strncmp(buf, "maxcopies=", 10) == 0)
				{
					p = buf;
					while(*p++ != '=');
					maxcopies = atoi(p);
				}
				if (strncmp(buf, "debug=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') debug = true;
					else debug = false;
				}
			}
		}
	}
	catch (...)
	{
		// plugin directory not set
		// probably using an older version avisynth
	}

    return new Dup(args[0].AsClip(),
		args[1].AsFloat(threshold),		// threshold for duplicate declaration
		args[2].AsBool(show),			// show biggest difference area
		args[3].AsBool(copy),			// copy
		args[4].AsInt(maxcopies),		// max successive copies to emit
		args[5].AsBool(debug),			// debug
		env);
}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env)
{
    env->AddFunction("Dup", "c[threshold]f[show]b[copy]b[maxcopies]i[debug]b", Create_Dup, 0);
    return 0;
}

